#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_exreporter
----------------------------------

Tests for `exreporter` module.
"""

import unittest
from mock import patch, MagicMock
from exreporter import exreporter
from exreporter.stores import Stores


class TestExreporter(unittest.TestCase):

    @patch('exreporter.exreporter.Reporter')
    def test_report_issue_with_defaults(self, Reporter):
        Reporter.return_value = reporter = MagicMock()
        credentials = object()
        store_name = 'some_store'

        exreporter.report_issue(credentials=credentials, store_name=store_name)

        reporter.report.assert_called_once_with(
            max_comments=50, include_locals=True, labels=['Bug'], time_delta=10)

    @patch('exreporter.exreporter.report_issue')
    def test_report_github_issue_with_defaults(self, report_issue):
        credentials = object()

        exreporter.report_github_issue(credentials=credentials)

        report_issue.assert_called_once_with(
            credentials=credentials, store_name='github')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
exporter.exporter
~~~~~~~~~~~~~~~~~

This module implements interface for Exporter library.
:copyright: (c) 2014 by Vedarth Kulkarni.
:license: MIT, see LICENSE for more details.

"""

from .reporter import Reporter


def report_issue(credentials, store_name, **kwargs):
    '''Reports an issue for any exception in application code to the specified store.

    :params credentials: object of store specific credentials from `exreporter.credentials`
    :params store_name: name of the issue store eg: 'github'
    :params title_format: (optional) string to specify the format of issue title
    :params body_format: (optional) string to specify the format of issue body
    :params extra_content: (optional) extra content that is to be added to issue body
    :params max_comments: (optional) maximum number of comments after which new issue is to be created, default value is ``50``
    :params time_delta: (optional) specifies minimum time interval after which issue should be reported if the exceptions occurs multiple times
    :params include_locals: (optional) boolean specifying whether dump of ``locals`` should be included in issue body
    :params labels: (optional) list of labels that are to be applied to the issue, default: ``['Bug']``
    '''
    kwargs.setdefault('max_comments', 50)
    kwargs.setdefault('time_delta', 10)
    kwargs.setdefault('include_locals', True)
    kwargs.setdefault('labels', ['Bug'])
    reporter = Reporter(credentials=credentials, store_name=store_name)
    return reporter.report(**kwargs)


def report_github_issue(credentials, **kwargs):
    '''Reports the exception to Github by calling ``report_issue``.

    :param credentials: object of :class:`GithubCrdentials` with correct github credentials
    :param \*\*kwargs: Optional arguments that ``report_issue`` takes.

    Returns :class:`GithubIssue` object.
    '''
    return report_issue(
        credentials=credentials, store_name='github', **kwargs)



# -*- coding: utf-8 -*-


class Formats(object):

    culprit = "`Culprit- {filepath}>{lineno}>{exception}`"

    title = "{exception} - {filename}::{method_name}"

    body = """
Following exception occurred:

```python
{stack_trace}
```
"""
    locals_format = """
Locals:
```python
{locals_data}
```

"""

    request_data = """
Request Data:

```python
{request_data}
```
"""



# -*- coding: utf-8 -*-

"""
test_reporter
----------------------------------

Tests for `exreporter.reporter` module.
"""

import unittest
from mock import patch, MagicMock

from exreporter.reporter import Reporter


class TestReporter(unittest.TestCase):

    @patch('exreporter.reporter.Formats')
    @patch('exreporter.reporter.StackTrace')
    @patch('exreporter.reporter.Stores')
    def test_reporter_report_with_defaults(self, Stores, StackTrace, Formats):
        Stores.some_name = store_cls = MagicMock()
        store = store_cls()
        store_name = 'some_name'
        credentials = object()
        stack_trace = StackTrace()
        stack_trace.exception = ValueError

        reporter = Reporter(credentials=credentials, store_name=store_name)

        reporter.report()

        store.create_or_update_issue.assert_called_once_with(
            title=Formats.title.format(),
            culprit=Formats.culprit.format(),
            body='''{}

{}
'''.format(Formats.body.format(), Formats.culprit.format()))



# -*- coding: utf-8 -*-

import sys
import traceback


class StackTrace(object):

    def __init__(self):
        ex_type, ex_value, ex_trace = sys.exc_info()

        assert ex_type is not None,\
            'No exception occurred, cannot proceed without any exception'

        filepath, method_name, locals_data, lineno = self._get_trace_info(
            trace=ex_trace)

        self.filepath = filepath
        self.locals_data = locals_data
        self.method_name = method_name
        self.lineno = lineno
        self.exception = ex_type
        self.traceback = ex_trace
        self.exception_value = ex_value

    def _get_trace_info(self, trace):
        custom_module_data = []

        while trace:
            filepath = trace.tb_frame.f_code.co_filename
            method_name = trace.tb_frame.f_code.co_name
            locals_data = trace.tb_frame.f_locals
            lineno = trace.tb_frame.f_lineno

            if 'site-packages' not in filepath:
                custom_module_data.append(
                    {
                        'filepath': filepath,
                        'method_name': method_name,
                        'locals_data': locals_data,
                        'lineno': lineno
                    }
                )
            trace = trace.tb_next

        if custom_module_data:
            data = custom_module_data.pop(-1)
            return (data['filepath'], data['method_name'],
                    data['locals_data'], data['lineno'])
        else:
            return (filepath, method_name, locals_data, lineno)

    @property
    def stack_trace_text(self):
        return '\n'.join(traceback.format_exception(
            self.exception, self.exception_value, self.traceback))



# -*- coding: utf-8 -*-

import os

from .stores import Stores
from .formats import Formats
from .stack_trace import StackTrace


class Reporter(object):

    def __init__(self, credentials, store_name):
        store_cls = getattr(Stores, store_name)
        self.store = store_cls(credentials=credentials)

    def report(self, **kwargs):
        trace_info = StackTrace()
        title_format = kwargs.pop('title_format', Formats.title)
        body_format = kwargs.pop('body_format', Formats.body)

        culprit = Formats.culprit.format(
            filepath=trace_info.filepath, lineno=trace_info.lineno,
            exception=trace_info.exception.__name__)

        title = title_format.format(
            exception=trace_info.exception.__name__,
            filename=os.path.basename(trace_info.filepath),
            method_name=trace_info.method_name)

        body = "{}".format(
            body_format.format(stack_trace=trace_info.stack_trace_text))

        if kwargs.pop('include_locals', False):
            body = "{} {}".format(
                body, Formats.locals_format.format(
                    locals_data=trace_info.locals_data)
            )

        extra_content = kwargs.pop('extra_content', '')

        if extra_content:
            body = "{}\n\nExtra Content:\n{}".format(
                body, extra_content
            )

        if kwargs.get('request'):
            body = "{} {}".format(
                body, Formats.request_data.format(
                    request_data=str(kwargs.get('request'))
                )
            )

        body = """{}

{}
""".format(body, culprit)

        return self.store.create_or_update_issue(
            title=title, body=body, culprit=culprit, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

from pip.req import parse_requirements

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
    'requests',
    'pytz',
    'python-dateutil',
]

test_requirements = requirements.extend([
    'mock',
    'magicmock',
])

setup(
    name='exreporter',
    version='0.1.1',
    description='Report Internal Server Errors as Issues',
    long_description=readme + '\n\n' + history,
    author='Vedarth Kulkarni',
    author_email='vedarthk@vedarthz.in',
    url='https://github.com/vedarthk/exreporter',
    packages=find_packages(),
    package_dir={'exreporter':
                 'exreporter'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='exreporter',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

"""
exporter.stores.github
~~~~~~~~~~~~~~~~~

This module implements Github issue store for exporter.
:copyright: (c) 2014 by Vedarth Kulkarni.
:license: MIT, see LICENSE for more details.

"""

import json
import pytz
import datetime
import requests
from dateutil.tz import tzlocal


class GithubCredentials(object):
    """Github credentials.

    Necessary credentials to access/create/update issues on Github.

    Basic Usage:

      >>> from exporter.stores.github import GithubCredentials
      >>> gc = GithubCredentials(user="username", repo="repo name", auth_token="personla auth token")
    """

    def __init__(self, user, repo, auth_token):
        self.user, self.repo, self.auth_token = user, repo, auth_token


class GithubStore(object):
    """Github Issue Store.
    """

    def __init__(self, credentials):
        '''Initializes Github issue store.

        :params credentials: object of :class:`GithubCredentials` with proper credentials
        '''
        assert type(credentials) is GithubCredentials,\
            'Credentials object is not of type GithubCredentials'
        self.credentials = credentials
        self.github_request = GithubRequest(credentials=credentials)

    def create_or_update_issue(self, title, body, culprit, **kwargs):
        '''Creates or comments on existing issue in the store.

        :params title: title for the issue
        :params body: body, the content of the issue
        :params culprit: string used to identify the cause of the issue,
            also used for aggregation
        :params labels: (optional) list of labels attached to the issue
        :returns: issue object
        :rtype: :class:`exreporter.stores.github.GithubIssue`
        '''
        issues = self.search(q=culprit)
        self.time_delta = kwargs.pop('time_delta')
        self.max_comments = kwargs.pop('max_comments')

        if issues:
            latest_issue = issues.pop(0)
            return self.handle_issue_comment(
                issue=latest_issue, title=title, body=body, **kwargs)
        else:
            return self.create_issue(
                title=title, body=body, **kwargs)

    def search(self, q, state='open,closed', **kwargs):
        """Search for issues in Github.

        :param q: query string to search
        :param state: state of the issue
        :returns: list of issue objects
        :rtype: list

        """
        search_result = self.github_request.search(q=q, state=state, **kwargs)
        if search_result['total_count'] > 0:
            return list(
                map(lambda issue_dict: GithubIssue(
                    github_request=self.github_request, **issue_dict),
                    search_result['items'])
            )

    def handle_issue_comment(self, issue, title, body, **kwargs):
        """Decides whether to comment or create a new issue when trying to comment.

        :param issue: issue on which the comment is to be added
        :param title: title of the issue if new one is to be created
        :param body: body of the issue/comment to be created
        :returns: newly created issue or the one on which comment was created
        :rtype: :class:`exreporter.stores.github.GithubIssue`

        """
        if self._is_time_delta_valid(issue.updated_time_delta):
            if issue.comments_count < self.max_comments:
                issue.comment(body=body)
                return issue
            else:
                return self.create_issue(title=title, body=body, **kwargs)

    def _is_time_delta_valid(self, delta):
        return delta > self.time_delta

    def create_issue(self, title, body, labels=None):
        """Creates a new issue in Github.

        :params title: title of the issue to be created
        :params body: body of the issue to be created
        :params labels: (optional) list of labels for the issue
        :returns: newly created issue
        :rtype: :class:`exreporter.stores.github.GithubIssue`
        """
        kwargs = self.github_request.create(
            title=title, body=body, labels=labels)
        return GithubIssue(github_request=self.github_request, **kwargs)


class GithubIssue(object):
    """Python object representation of issue on Github.
    """

    def __init__(self, github_request, **kwargs):
        """Initializes Github issue object.
        For **kwargs please refer Github's documentation:
        https://developer.github.com/v3/issues/#get-a-single-issue

        :params github_request: object of :class:`GithubRequest` used to make HTTP requests to Github
        """
        self.github_request = github_request
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @property
    def comments_count(self):
        """Returns number of comments on the issue.
        """
        return int(self.comments)

    @property
    def updated_time_delta(self):
        """Returns the number of seconds ago the issue was updated from current time.
        """
        local_timezone = tzlocal()
        update_at = datetime.datetime.strptime(self.updated_at, '%Y-%m-%dT%XZ')
        update_at_utc = pytz.utc.localize(update_at)
        update_at_local = update_at_utc.astimezone(local_timezone)
        delta = datetime.datetime.now(local_timezone) - update_at_local
        return int(delta.total_seconds())

    def open_issue(self):
        """Changes the state of issue to 'open'.
        """
        self.github_request.update(issue=self, state='open')
        self.state = 'open'

    def comment(self, body):
        """Adds a comment to the issue.

        :params body: body, content of the comment
        :returns: issue object
        :rtype: :class:`exreporter.stores.github.GithubIssue`
        """
        self.github_request.comment(issue=self, body=body)

        if self.state == 'closed':
            self.open_issue()
        return self


class GithubRequest(object):
    """This class objects are created with valid credentials.
    The objects have methods to create/update issues and to add comments on Github.

    Basic Usage:
      >>> from exreporter.stores.github import GithubCredentials, GithubRequest
      >>> gc = GithubCredentials(user="username", repo="repo name", auth_token="personla auth token")
      >>> gr = GithubRequest(credentials=gc)
      >>> gr.create(title="title", body="body", labels=['labels'])
    """

    def __init__(self, credentials):
        self.user = credentials.user
        self.repo = credentials.repo
        self.session = requests.Session()
        self.session.headers.update({
            'Authorization': 'token {}'.format(credentials.auth_token)
        })

    def create(self, title, body, labels):
        """Create an issue in Github.
        For JSON data returned by Github refer:
        https://developer.github.com/v3/issues/#create-an-issue

        :param title: title of the issue
        :param body: body of the issue
        :param labels: list of labels for the issue
        :returns: dict of JSON data returned by Github of newly created issue
        :rtype: `dict`

        """
        url = "https://api.github.com/repos/{}/{}/issues".format(
            self.user, self.repo)

        data = {
            'title': title,
            'body': body,
        }

        if labels:
            data.update({'labels': labels})

        response = self.session.post(url, json.dumps(data))

        assert response.status_code == 201
        return json.loads(response.content)

    def comment(self, issue, body):
        """Comment on existing issue on Github.
        For JSON data returned by Github refer:
        https://developer.github.com/v3/issues/comments/#create-a-comment

        :param issue: object of exisiting issue
        :param body: body of the comment
        :returns: dict of JSON data returned by Github of the new comment
        :rtype: `dict`

        """
        url = issue.comments_url
        data = {'body': body}

        response = self.session.post(url, json.dumps(data))

        assert response.status_code == 201
        return json.loads(response.content)

    def update(self, issue, **kwargs):
        """Update an existing issue on Github.
        For JSON data returned by Github refer:
        https://developer.github.com/v3/issues/#edit-an-issue

        :param issue: object existing issue
        :returns: dict of JSON data returned by Github.
        :rtype: `dict`

        """
        url = issue.url

        response = self.session.patch(url, json.dumps(kwargs))

        assert response.status_code == 200
        return json.loads(response.content)

    def search(self, q, state, labels=None):
        """Search for issues in Github.
        For JSON data returned by Github refer:
        https://developer.github.com/v3/search/#search-issues

        :param q: query string for search
        :param state: the states of the issue
        :param labels: labels of the issue
        :returns: dictionary of JSON data returned by Github
        :rtype: `dict`

        """
        # TODO: add support for search with labels
        q = "{}+state:{}".format(q, '+state:'.join(state.split(',')))
        sort = "updated"

        url = "https://api.github.com/search/"\
              "issues?q={}+repo:{}/{}&sort={}".format(
                  q, self.user, self.repo, sort)

        response = self.session.get(url)
        assert response.status_code == 200
        return json.loads(response.content)



# -*- coding: utf-8 -*-

"""
exporter.contrib.django_middlewares
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements Django middlewares for different stores.
Requires following Django settings to be present:
- EXREPORTER_GITHUB_USER
- EXREPORTER_GITHUB_REPO
- EXREPORTER_GITHUB_AUTH_TOKEN
- EXREPORTER_GITHUB_LABELS

:copyright: (c) 2014 by Vedarth Kulkarni.
:license: MIT, see LICENSE for more details.

"""

from django.conf import settings
from exreporter.credentials import GithubCredentials
from exreporter import exreporter


class ExreporterGithubMiddleware(object):
    """Exreporter middleware for Django framework.
    """

    def process_exception(self, request, exception):
        """Report exceptions from requests via Exreporter.
        """
        credentials = GithubCredentials(
            user=settings.EXREPORTER_GITHUB_USER,
            repo=settings.EXREPORTER_GITHUB_REPO,
            auth_token=settings.EXREPORTER_GITHUB_AUTH_TOKEN)

        exreporter.report_github_issue(
            credentials=credentials, labels=settings.EXREPORTER_GITHUB_LABELS)
